﻿using InfluxClient = InfluxDB.Client;
using Serilog;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Threading.Tasks;

namespace BenchmarkTool
{
    class DatabaseWrapper : IDatabase
    {
        private IDatabase TargetDatabase;
        private IMetric performanceMetric;

        public DatabaseWrapper(IMetric metric)
        {
            try
            {
                DatabaseFactory dbFactory = new DatabaseFactory();
                TargetDatabase = dbFactory.Create();
            }
            catch (Exception ex)
            {
                Log.Error(ex.ToString());
            }
            performanceMetric = metric;
        }        

        public void cleanup()
        {
            TargetDatabase.cleanup();
        }

        public void close()
        {
            TargetDatabase.close();
        }

        public void init()
        {
            TargetDatabase.init();
        }

        public QueryStatus OutOfRangeQuery(OORangeQuery query)
        {
            QueryStatus status = null;
            var operation = Operation.OutOfRangeQuery;
            try
            {
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                status = TargetDatabase.OutOfRangeQuery(query);
                stopwatch.Stop();
                status.Latency = stopwatch.ElapsedMilliseconds;
                if (status.Succeeded)
                {
                    measureOkOperation(status, operation, status.DataPoints);
                }
                else
                {
                    performanceMetric.IncrementFailedOperations( 1);
                    performanceMetric.IncrementFailedPoints( status.DataPoints);
                }
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute Out of range query: {0}", ex.ToString()));
                performanceMetric.IncrementFailedOperations( 1);
                performanceMetric.IncrementFailedPoints( status.DataPoints);
            }
            return status;
        }


        public QueryStatus AggregatedDifferenceQuery(ComparisionQuery query)
        {
            QueryStatus status = null;
            var operation = Operation.RangeQueryAggData;
            try
            {
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                status = TargetDatabase.AggregatedDifferenceQuery(query);
                stopwatch.Stop();
                status.Latency = stopwatch.ElapsedMilliseconds;
                if (status.Succeeded)
                {
                    measureOkOperation(status, operation, status.DataPoints);
                }
                else
                {
                    performanceMetric.IncrementFailedOperations( 1);
                    performanceMetric.IncrementFailedPoints( status.DataPoints);
                }
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute aggregated difference range query: {0}", ex.ToString()));
                performanceMetric.IncrementFailedOperations( 1);
                performanceMetric.IncrementFailedPoints( status.DataPoints);
            }
            return status;
        }

        public QueryStatus RangeQueryAgg(RangeQuery rangeQuery)
        {
            QueryStatus status = null;
            var operation = Operation.RangeQueryAggData;
            try
            {
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                status = TargetDatabase.RangeQueryAgg(rangeQuery);
                stopwatch.Stop();
                status.Latency = stopwatch.ElapsedMilliseconds;
                if (status.Succeeded)
                {
                    measureOkOperation(status, operation, status.DataPoints);
                }
                else
                {
                    performanceMetric.IncrementFailedOperations( 1);
                    performanceMetric.IncrementFailedPoints( status.DataPoints);
                }
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute agg range query: {0}", ex.ToString()));
                performanceMetric.IncrementFailedOperations( 1);
                performanceMetric.IncrementFailedPoints( status.DataPoints);
            }
            return status;
        }

        public QueryStatus RangeQueryRaw(RangeQuery rangeQuery)
        {
            QueryStatus status = null;
            var operation = Operation.RangeQueryRawData;
            try
            {
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                status = TargetDatabase.RangeQueryRaw(rangeQuery);
                stopwatch.Stop();
                status.Latency = stopwatch.ElapsedMilliseconds;
                Log.Information(stopwatch.ElapsedMilliseconds.ToString());
                if (status.Succeeded)
                {
                    Log.Information("Succeeded");
                    measureOkOperation(status, operation, status.DataPoints);
                    Log.Information(String.Format("Data points: {0}", status.DataPoints.ToString()));

                }
                else
                {
                    performanceMetric.IncrementFailedOperations( 1);
                    performanceMetric.IncrementFailedPoints( status.DataPoints);
                }
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute range query: {0}", ex.ToString()));
                performanceMetric.IncrementFailedOperations( 1);
                performanceMetric.IncrementFailedPoints(status.DataPoints);
            }
            return status;
        }

        public QueryStatus StandardDevQuery(SpecificQuery query)
        {
            QueryStatus status = null;
            var operation = Operation.STDDevQuery;
            try
            {
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                status = TargetDatabase.StandardDevQuery(query);
                stopwatch.Stop();
                status.Latency = stopwatch.ElapsedMilliseconds;
                if (status.Succeeded)
                {
                    measureOkOperation(status, operation, status.DataPoints);
                }
                else
                {
                    performanceMetric.IncrementFailedOperations( 1);
                    performanceMetric.IncrementFailedPoints( status.DataPoints);
                }
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute agg range query: {0}", ex.ToString()));
                performanceMetric.IncrementFailedOperations( 1);
                performanceMetric.IncrementFailedPoints( status.DataPoints);
            }
            return status;
        }

        //public QueryStatus WriteBatch(Batch batch)
        //{
        //    var operation = Operation.Ingestion;
        //    QueryStatus status = null;
        //    try
        //    {
        //        Stopwatch stopwatch = new Stopwatch();
        //        stopwatch.Start();
        //        status = TargetDatabase.WriteBatch(batch);
        //        stopwatch.Stop();
        //        status.Latency = stopwatch.ElapsedMilliseconds;
        //        if (status.Succeeded)
        //        {
        //            measureOkOperation(status, operation, batch.Size);
        //        }
        //        else
        //        {
        //            performanceMetric.IncrementFailedOperations(operation, 1);
        //            performanceMetric.IncrementFailedPoints(operation, batch.Size);
        //        }
        //    }
        //    catch (Exception ex)
        //    {
        //        Log.Error(String.Format("Failed to insert batch. {0}", ex.ToString()));
        //        performanceMetric.IncrementFailedOperations(operation, 1);
        //        performanceMetric.IncrementFailedPoints(operation, batch.Size);
        //    }
        //    return status;
        //}


        public async Task<QueryStatus> WriteBatch(Batch batch)
        {
            var operation = Operation.Ingestion;
            QueryStatus status = null;
            try
            {
                Stopwatch stopwatch = new Stopwatch();
                stopwatch.Start();
                status = await TargetDatabase.WriteBatch(batch);

                stopwatch.Stop();
                status.Latency = stopwatch.ElapsedMilliseconds;

                if (status.Succeeded)
                {
                    measureOkOperation(status, operation, batch.Size);
                }
                else
                {
                    performanceMetric.IncrementFailedOperations( 1);
                    performanceMetric.IncrementFailedPoints( batch.Size);
                }
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to insert batch. {0}", ex.ToString()));
                performanceMetric.IncrementFailedOperations( 1);
                performanceMetric.IncrementFailedPoints( batch.Size);
            }
            return status;
        }

        private void measureOkOperation(QueryStatus status, Operation operation, int pointsNumber)
        {
            var latency = status.Latency;
            if (latency < 0)
            {
                Log.Error(String.Format("Negative latency for operation: {0}. This might be caused by an error", operation.ToString()));
                latency = 0;
            }
            performanceMetric.IncrementOperationLatency( latency);
            performanceMetric.IncrementSucceededOperations( 1);
            performanceMetric.IncrementSucceededPoints( pointsNumber);
            performanceMetric.ExecutionTime = latency;
        }

        
    }
}