﻿using InfluxDB.Client;
using InfluxDB.Client.Api.Domain;
using Serilog;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using System.Numerics;

namespace BenchmarkTool
{
    public class InfluxDB : IDatabase
    {
        private static readonly DateTime EpochStart = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc);
        private InfluxDBClientOptions options;
        private InfluxDBClient client;
        private WriteApiAsync writeApi ;


        public void cleanup()
        {
            throw new NotImplementedException();
        }

        public void close()
        {
            try
            {
                if (client != null)
                    client.Dispose();
                
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to close InfluxDB. Exception: {0}", ex.ToString()));
            }
        }

        public void init()
        {
            try
            {
                TimeSpan t = new TimeSpan(0, 0, 10,0);
                options = new InfluxDBClientOptions.Builder()
                    .Url(Config.GetInfluxHost())
                    .AuthenticateToken(Config.GetInfluxToken().ToCharArray())
                    .Org(Config.GetInfluxOrganization())
                    .Bucket(Config.GetInfluxBucket())
                    .ReadWriteTimeOut(t)
                    .Build();
                client = InfluxDBClientFactory.Create(options);
                
                writeApi = client.GetWriteApiAsync();
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to initialize InfluxDB. Exception: {0}", ex.ToString()));
            }
        }

        public QueryStatus OutOfRangeQuery(OORangeQuery query)
        {

            try
            {

                var flux = Query.FluxOutOfRangeQuery.Replace(QueryParams.StartParam, query.StartDate.ToUniversalTime().ToString("o"));
                flux = flux.Replace(QueryParams.EndParam, query.EndDate.ToUniversalTime().ToString("o"));
                flux = flux.Replace(QueryParams.SensorIDParam, query.SensorID.ToString());
                flux = flux.Replace(QueryParams.MaxValParam, query.MaxValue.ToString());
                flux = flux.Replace(QueryParams.MinValParam, query.MinValue.ToString());
               // Log.Information(String.Format("Flux query: {0}", flux));

                var queryApi = client.GetQueryApiSync();
                var results = queryApi.QuerySync(flux, Config.GetInfluxOrganization());
                Log.Information(String.Format("Number of point: {0}", results[0].Records.Count.ToString()));
                return new QueryStatus(true, results != null && results.Count > 0 ? results[0].Records.Count : 0);
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute Out of Range Query on InfluxDB. Exception: {0}", ex.ToString()));
                return new QueryStatus(false, 0, ex, ex.ToString());
            }
        }

        public QueryStatus RangeQueryAgg(RangeQuery rangeQuery)
        {
            try
            {

                var flux = Query.FluxRangeQueryAgg.Replace(QueryParams.StartParam, rangeQuery.StartDate.ToUniversalTime().ToString("o"));
                flux = flux.Replace(QueryParams.EndParam, rangeQuery.EndDate.ToUniversalTime().ToString("o"));
                var sensorIds = rangeQuery.SensorIDs.Select(x => String.Concat("^", x, "$")).ToList();
                var ids = String.Concat("/", String.Join("|", sensorIds), "/");
                flux = flux.Replace(QueryParams.SensorIDsParam, ids);
                Log.Information(String.Format("Flux query: {0}", flux));

                var queryApi = client.GetQueryApiSync();
                var results = queryApi.QuerySync(flux, Config.GetInfluxOrganization());
                Log.Information(String.Format("Number of point: {0}", results[0].Records.Count.ToString()));
                return new QueryStatus(true, results != null && results.Count > 0 ? results[0].Records.Count : 0);
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute Range Query Aggregated Data on InfluxDB. Exception: {0}", ex.ToString()));
                return new QueryStatus(false, 0, ex, ex.ToString());
            }
        }

        public QueryStatus RangeQueryRaw(RangeQuery rangeQuery)
        {
            try
            {

                var flux = Query.FluxRangeQueryRaw.Replace(QueryParams.StartParam, rangeQuery.StartDate.ToUniversalTime().ToString("o"));
                flux = flux.Replace(QueryParams.EndParam, rangeQuery.EndDate.ToUniversalTime().ToString("o"));
                var sensorIds = rangeQuery.SensorIDs.Select(x => String.Concat("^", x , "$")).ToList();
                var ids = String.Concat("/", String.Join("|", sensorIds), "/");
                flux = flux.Replace(QueryParams.SensorIDsParam, ids);
                Log.Information(String.Format("Flux query: {0}", flux));

                var queryApi = client.GetQueryApiSync();
                var results = queryApi.QuerySync(flux, Config.GetInfluxOrganization());                
                Log.Information(String.Format("Number of point: {0}", results[0].Records.Count.ToString()));
                return new QueryStatus(true, results != null && results.Count > 0?  results[0].Records.Count : 0);
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute Range Query Raw Data on InfluxDB. Exception: {0}", ex.ToString()));
                return new QueryStatus(false, 0, ex, ex.ToString());
            }
        }

        public QueryStatus AggregatedDifferenceQuery(ComparisionQuery query)
        {
            try
            {

                var flux = Query.FluxAggDifferenceQuery.Replace(QueryParams.StartParam, query.StartDate.ToUniversalTime().ToString("o"));
                flux = flux.Replace(QueryParams.EndParam, query.EndDate.ToUniversalTime().ToString("o"));
                flux = flux.Replace(QueryParams.FirstSensorIDParam, query.FirstSensorID.ToString());
                flux = flux.Replace(QueryParams.SecondSensorIDParam, query.SecondSensorID.ToString());
                Log.Information(String.Format("Flux query: {0}", flux));

                var queryApi = client.GetQueryApiSync();
                var results = queryApi.QuerySync(flux, Config.GetInfluxOrganization());
                Log.Information(String.Format("Number of point: {0}", results[0].Records.Count.ToString()));
                return new QueryStatus(true, results != null && results.Count > 0 ? results[0].Records.Count : 0);
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute Difference beween agg sensor values query on InfluxDB. Exception: {0}", ex.ToString()));
                return new QueryStatus(false, 0, ex, ex.ToString());
            }
        }


        async Task<QueryStatus> IDatabase.WriteBatch(Batch batch)
        {
            try
            {
                await client.GetWriteApiAsync().WriteMeasurementsAsync(WritePrecision.Ns, batch.Records);
                return new QueryStatus(true);
                
                //Replace with the below code to optimize InfluxDB client lib 
                //var sb = new StringBuilder("");

                //foreach (var item in batch.Records)
                //{
                //    var timeSpan = item.Time.Subtract(EpochStart);
                //    var time = TimeSpanToBigInteger(timeSpan, WritePrecision.Ns);

                //    sb.Append(String.Format("{0},sensor_id={1} value={2} {3}",Constants.DatabaseName, item.SensorID.ToString(), item.Value.ToString(), time.ToString()));
                //    sb.Append("\n");
                //}
                //if (sb.Length != 0)
                //{
                //    // remove last \n
                //    sb.Remove(sb.Length - 1, 1);
                //}

                //await writeApi._service.PostWriteAsync(Config.GetInfluxOrganization(), Config.GetInfluxBucket(), Encoding.UTF8.GetBytes(sb.ToString()), null,
                //    "identity", "text /plain; charset=utf-8", null, "application /json", null, InfluxClient.Api.Domain.WritePrecision.Ns,
                //    default);

                return new QueryStatus(true);
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to insert batch into InfluxDB. Exception: {0}", ex.ToString()));
                return new QueryStatus(false, 0, ex, ex.ToString());
            }
        }
        public QueryStatus StandardDevQuery(SpecificQuery query)
        {
            try
            {
                var flux = Query.FluxSTDEVQuery.Replace(QueryParams.StartParam, query.StartDate.ToUniversalTime().ToString("o"));
                flux = flux.Replace(QueryParams.EndParam, query.EndDate.ToUniversalTime().ToString("o"));               
                flux = flux.Replace(QueryParams.SensorIDParam, query.SensorID.ToString());
                Log.Information(String.Format("Flux query: {0}", flux));

                var queryApi = client.GetQueryApiSync();
                var results = queryApi.QuerySync(flux, Config.GetInfluxOrganization());
                Log.Information(String.Format("Number of point: {0}", results[0].Records.Count.ToString()));
                return new QueryStatus(true, results != null && results.Count > 0 ? results[0].Records.Count : 0);
            }
            catch (Exception ex)
            {
                Log.Error(String.Format("Failed to execute STD Dev query on Influxdb. Exception: {0}", ex.ToString()));
                return new QueryStatus(false, 0, ex, ex.ToString());
            }
        }

        private static BigInteger TimeSpanToBigInteger(TimeSpan timestamp, WritePrecision timeUnit)
        {
            BigInteger time;
            switch (timeUnit)
            {
                case WritePrecision.Ns:
                    time = timestamp.Ticks * 100;
                    break;
                case WritePrecision.Us:
                    time = (BigInteger)(timestamp.Ticks * 0.1);
                    break;
                case WritePrecision.Ms:
                    time = (BigInteger)timestamp.TotalMilliseconds;
                    break;
                case WritePrecision.S:
                    time = (BigInteger)timestamp.TotalSeconds;
                    break;
                default:
                    throw new ArgumentOutOfRangeException(nameof(timeUnit), timeUnit, "WritePrecision value is not supported");
            }

            return time;
        }
    }
}
    

