﻿using PostgreSQLCopyHelper;
using System;
using Npgsql;
using System.Collections.Generic;
using System.Threading.Tasks;
using System.Diagnostics;
using Serilog;
using Serilog.Events;
using System.Collections.Concurrent;
using System.Linq;

namespace BenchmarkTool
{
    static class Program
    {
        static async Task Main(string[] args)
        {
            try
            {
                Log.Logger = new LoggerConfiguration()
                .MinimumLevel.Debug()
                .WriteTo.File("log.txt", restrictedToMinimumLevel: LogEventLevel.Information)
                .CreateLogger();

                DatabaseFactory dbFactory = new DatabaseFactory();
                Console.WriteLine("Starting...");
                Log.Information("Application started");
                var action = "Ingestion";
                if (args != null && args.Length > 0)
                {
                    action = args[0];
                }

                switch (action)
                {
                    case "PopulateData":
                        await PopulateData();
                        break;

                    case "BenchmarkReadData":
                        await BenchmarkReadData();
                        break;

                    case "BenchmarkHotReadData":
                        await BenchmarkHotReadData();
                        break;
                    case "Ingestion":
                        int clientsNb = 0;
                        int sensorsNb = Config.GetSensorNumber();
                        int batchSize = 0;
                        int[] clientNumberArray = Config.GetClientNumberOptions();

                        int[] batchSizeArray = Config.GetBatchSizeOptions();
                        int l = 0;
                        for (var c = 0; c < clientNumberArray.Length; c++)
                        {
                            clientsNb = clientNumberArray[c];
                            for (var b = 0; b < batchSizeArray.Length; b++)
                            {
                                DateTime date = Config.GetStartTime().AddMinutes(Config.GetGlobalTimeOffset() * l * Config.GetTestLoop());
                                batchSize = batchSizeArray[b];
                                for (var loop = 0; loop < Config.GetTestLoop(); loop++)
                                {
                                    List<Client> clients = new List<Client>();

                                    if (clientsNb > 1)
                                    {
                                        Stopwatch stopwatch = new Stopwatch();
                                        stopwatch.Start();
                                        for (var i = 0; i < clientsNb; i++)
                                        {
                                            clients.Add(new Client(i, clientsNb, batchSize, sensorsNb, date, loop));
                                        }
                                        await clients.ParallelForEachAsync(RunQueryTask, clientsNb);
                                        stopwatch.Stop();
                                        var metric = CalculateOverallPerformance(clients, stopwatch.ElapsedMilliseconds, sensorsNb, clientsNb, batchSize);
                                        metric.WriteToCSV();
                                    }
                                    else
                                    {
                                        var client = new Client(0, 1, batchSize, sensorsNb, date, loop);
                                        await client.RunQuery();
                                        clients.Add(client);
                                        client.PerformanceMetric.WriteToCSV();
                                    }
                                    l++;
                                }
                            }
                        }
                        break;
                }
                Log.Information("Completed..");
                Console.WriteLine("Completed...");
            }
            catch (Exception ex)
            {
                Log.Error(ex.ToString());
            }
        }

        private static async Task PopulateData()
        {
            DatabaseFactory dbFactory = new DatabaseFactory();
            Console.WriteLine("Starting...");
            Log.Information("Application started");
            var days = Config.GetDaySpan();
            DateTime date = DateTime.UtcNow.AddDays(-days); 
            var batchSize = Config.GetBatchSize();
            var loop = Config.GetTestLoop();
            var sensorsNb = Config.GetSensorNumber();
            //Loop through 30 days to populate the database
            for (var d = 0; d < days; d++)
            {
                for (int l = 0; l < loop; l++)
                {
                    var client = new Client(0, 1, batchSize, sensorsNb, date, l);
                    await client.RunQuery();
                }
                date = date.AddDays(1);
            }
        }

        private static async Task BenchmarkReadData()
        {
            var clientsNumber = Config.GetClientNumber();
            Random rnd = new Random();
            var duration = Config.GetDurationMinutes();
            DateTime startDate = Config.GetStartTime().AddMinutes(-duration);
            var iterations = Config.GetTestLoop();
            var dataSpan = Config.GetDaySpan() - 1;
            for (var loop = 0; loop < iterations; loop++)
            {
                var client = new ClientRead(clientsNumber, startDate.AddDays(-rnd.Next(dataSpan)));
                await client.RunQuery();
                client.PerformanceMetric.WriteToCSV();
            }
        }


        private static async Task BenchmarkHotReadData()
        {
            var clientsNumber = Config.GetClientNumber();
            var span = Config.GetHourSpan();
            var duration = Config.GetDurationMinutes();
            DateTime startDate = Config.GetStartTime();
            var iterations = Config.GetTestLoop();
            for (var loop = 0; loop < iterations; loop++)
            {
                var client = new ClientRead(clientsNumber, startDate.AddHours(loop % span));
                await client.RunQuery();
                client.PerformanceMetric.WriteToCSV();
            }
        }

        static PerformanceMetricWrite CalculateOverallPerformance(List<Client> clients, long elapsedTime, int sensorNb, int clientNb, int batchSize)
        {
            var overallMetric = new PerformanceMetricWrite(batchSize, clientNb, sensorNb);
            overallMetric.ExecutionTime = elapsedTime;
            foreach (var client in clients)
            {
                overallMetric.Merge(client.PerformanceMetric);
            }
            return overallMetric;
        }

        public static Task ParallelForEachAsync<Client>(this IEnumerable<Client> source, Func<Client, Task> func, int maxDoP = 4)
        {
            async Task AwaitPartition(IEnumerator<Client> partition)
            {
                using (partition)
                {
                    while (partition.MoveNext())
                    {
                        await Task.Yield(); // prevents a sync/hot thread hangup
                        await func(partition.Current);
                    }
                }
            }

            return Task.WhenAll(
                Partitioner
                    .Create(source)
                    .GetPartitions(maxDoP)
                    .AsParallel()
                    .Select(p => AwaitPartition(p)));
        }

        public static async Task RunQueryTask(Client client)
        {
            await client.RunQuery();
        }
    }
}

